"""Configuration module."""

import os
import socket
import getpass

from omegaconf import OmegaConf
import random

########################
#   Project Settings   #
########################

PROJECT_NAME = os.getenv("PROJECT_NAME", "flowcept")
_SETTINGS_DIR = os.path.expanduser(f"~/.{PROJECT_NAME}")
SETTINGS_PATH = os.getenv("FLOWCEPT_SETTINGS_PATH", f"{_SETTINGS_DIR}/settings.yaml")

if not os.path.exists(SETTINGS_PATH):
    SETTINGS_PATH = None
    from importlib import resources

    with resources.files("resources").joinpath("sample_settings.yaml").open("r") as f:
        settings = OmegaConf.load(f)
else:
    settings = OmegaConf.load(SETTINGS_PATH)

########################
#   Log Settings       #
########################

LOG_FILE_PATH = settings["log"].get("log_path", "default")

if LOG_FILE_PATH == "default":
    LOG_FILE_PATH = f"{PROJECT_NAME}.log"

# Possible values below are the typical python logging levels.
LOG_FILE_LEVEL = settings["log"].get("log_file_level", "disable").upper()
LOG_STREAM_LEVEL = settings["log"].get("log_stream_level", "disable").upper()

##########################
#  Experiment Settings   #
##########################

FLOWCEPT_USER = settings["experiment"].get("user", "blank_user")
CAMPAIGN_ID = settings["experiment"].get(
    "campaign_id", os.environ.get("CAMPAIGN_ID", "super_campaign")
)

######################
#   MQ Settings   #
######################

MQ_URI = settings["mq"].get("uri", None)
MQ_INSTANCES = settings["mq"].get("instances", None)

MQ_TYPE = os.getenv("MQ_TYPE", settings["mq"].get("type", "redis"))
MQ_CHANNEL = settings["mq"].get("channel", "interception")
MQ_PASSWORD = settings["mq"].get("password", None)
MQ_HOST = os.getenv("MQ_HOST", settings["mq"].get("host", "localhost"))
MQ_PORT = int(os.getenv("MQ_PORT", settings["mq"].get("port", "6379")))

MQ_BUFFER_SIZE = int(settings["mq"].get("buffer_size", 50))
MQ_INSERTION_BUFFER_TIME = int(settings["mq"].get("insertion_buffer_time_secs", 5))
MQ_INSERTION_BUFFER_TIME = random.randint(
    int(MQ_INSERTION_BUFFER_TIME * 0.9),
    int(MQ_INSERTION_BUFFER_TIME * 1.4),
)
MQ_CHUNK_SIZE = int(settings["mq"].get("chunk_size", -1))

#####################
# KV SETTINGS       #
#####################

KVDB_PASSWORD = settings["kv_db"].get("password", None)
KVDB_HOST = os.getenv("KVDB_HOST", settings["kv_db"].get("host", "localhost"))
KVDB_PORT = int(os.getenv("KVDB_PORT", settings["kv_db"].get("port", "6379")))

DATABASES = settings.get("databases", {})

######################
#  MongoDB Settings  #
######################
_mongo_settings = DATABASES.get("mongodb", None)
MONGO_ENABLED = False
if _mongo_settings:
    if "MONGO_ENABLED" in os.environ:
        MONGO_ENABLED = os.environ.get("MONGO_ENABLED").lower() == "true"
    else:
        MONGO_ENABLED = _mongo_settings.get("enabled", False)
    MONGO_URI = os.environ.get("MONGO_URI") or _mongo_settings.get("uri")
    MONGO_HOST = os.environ.get("MONGO_HOST") or _mongo_settings.get("host", "localhost")
    MONGO_PORT = int(os.environ.get("MONGO_PORT") or _mongo_settings.get("port", 27017))
    MONGO_DB = _mongo_settings.get("db", PROJECT_NAME)
    MONGO_CREATE_INDEX = _mongo_settings.get("create_collection_index", True)

######################
#  LMDB Settings  #
######################
_lmdb_settings = DATABASES.get("lmdb", None)
LMDB_ENABLED = False
if _lmdb_settings:
    if "LMDB_ENABLED" in os.environ:
        LMDB_ENABLED = os.environ.get("LMDB_ENABLED").lower() == "true"
    else:
        LMDB_ENABLED = _lmdb_settings.get("enabled", False)

##########################
# Buffer Settings        #
##########################
_buffer_settings = settings["buffer"]
# In seconds:
INSERTION_BUFFER_TIME = int(_buffer_settings.get("insertion_buffer_time_secs", 5))
INSERTION_BUFFER_TIME = random.randint(
    int(INSERTION_BUFFER_TIME * 0.9),
    int(INSERTION_BUFFER_TIME * 1.4),
)

ADAPTIVE_BUFFER_SIZE = _buffer_settings.get("adaptive_buffer_size", True)
MAX_BUFFER_SIZE = int(_buffer_settings.get("max_buffer_size", 50))
MIN_BUFFER_SIZE = max(1, int(_buffer_settings.get("min_buffer_size", 10)))
REMOVE_EMPTY_FIELDS = _buffer_settings.get("remove_empty_fields", False)

######################
# PROJECT SYSTEM SETTINGS #
######################

DB_FLUSH_MODE = settings["project"].get("db_flush_mode", "online")
# DEBUG_MODE = settings["project"].get("debug", False)
PERF_LOG = settings["project"].get("performance_logging", False)
JSON_SERIALIZER = settings["project"].get("json_serializer", "default")
REPLACE_NON_JSON_SERIALIZABLE = settings["project"].get("replace_non_json_serializable", True)
ENRICH_MESSAGES = settings["project"].get("enrich_messages", True)
REGISTER_WORKFLOW = settings["project"].get("register_workflow", True)

TELEMETRY_CAPTURE = settings.get("telemetry_capture", None)

##################################
# GPU TELEMETRY CAPTURE SETTINGS #
#################################

#  TODO: This is legacy. We should improve the way to set these
#   initial variables and initialize GPU libs.
#   We could move this to the static part of TelemetryCapture
N_GPUS = dict()
GPU_HANDLES = None
if TELEMETRY_CAPTURE is not None and TELEMETRY_CAPTURE.get("gpu", None) is not None:
    if eval(TELEMETRY_CAPTURE.get("gpu", "None")) is not None:
        try:
            visible_devices_var = os.environ.get("CUDA_VISIBLE_DEVICES", None)
            if visible_devices_var is not None:
                visible_devices = [int(i) for i in visible_devices_var.split(",")]
                if len(visible_devices):
                    N_GPUS["nvidia"] = visible_devices
                    GPU_HANDLES = []  # TODO
            else:
                from pynvml import nvmlDeviceGetCount

                N_GPUS["nvidia"] = list(range(0, nvmlDeviceGetCount()))
                GPU_HANDLES = []
        except Exception:
            pass
        try:
            visible_devices_var = os.environ.get("ROCR_VISIBLE_DEVICES", None)
            if visible_devices_var is not None:
                visible_devices = [int(i) for i in visible_devices_var.split(",")]
                if len(visible_devices):
                    N_GPUS["amd"] = visible_devices
                    from amdsmi import (
                        amdsmi_init,
                        amdsmi_get_processor_handles,
                    )

                    amdsmi_init()
                    GPU_HANDLES = amdsmi_get_processor_handles()
            else:
                from amdsmi import amdsmi_init, amdsmi_get_processor_handles

                amdsmi_init()
                GPU_HANDLES = amdsmi_get_processor_handles()
                N_GPUS["amd"] = list(range(0, len(GPU_HANDLES)))
        except Exception:
            pass

if len(N_GPUS.get("amd", [])):
    GPU_TYPE = "amd"
elif len(N_GPUS.get("nvidia", [])):
    GPU_TYPE = "nvidia"
else:
    GPU_TYPE = None

######################
# SYS METADATA #
######################

LOGIN_NAME = None
PUBLIC_IP = None
PRIVATE_IP = None
SYS_NAME = None
NODE_NAME = None
ENVIRONMENT_ID = None

sys_metadata = settings.get("sys_metadata", None)
if sys_metadata is not None:
    ENVIRONMENT_ID = sys_metadata.get("environment_id", None)
    SYS_NAME = sys_metadata.get("sys_name", None)
    NODE_NAME = sys_metadata.get("node_name", None)
    LOGIN_NAME = sys_metadata.get("login_name", None)
    PUBLIC_IP = sys_metadata.get("public_ip", None)
    PRIVATE_IP = sys_metadata.get("private_ip", None)


if LOGIN_NAME is None:
    try:
        LOGIN_NAME = sys_metadata.get("login_name", getpass.getuser())
    except Exception:
        try:
            LOGIN_NAME = os.getlogin()
        except Exception:
            LOGIN_NAME = None

SYS_NAME = SYS_NAME if SYS_NAME is not None else os.uname()[0]
NODE_NAME = NODE_NAME if NODE_NAME is not None else os.uname()[1]

try:
    HOSTNAME = socket.getfqdn()
except Exception:
    try:
        HOSTNAME = socket.gethostname()
    except Exception:
        try:
            with open("/etc/hostname", "r") as f:
                HOSTNAME = f.read().strip()
        except Exception:
            HOSTNAME = "unknown_hostname"


EXTRA_METADATA = settings.get("extra_metadata", {})
EXTRA_METADATA.update({"mq_host": MQ_HOST})
EXTRA_METADATA.update({"mq_port": MQ_PORT})

######################
#    Web Server      #
######################
settings.setdefault("web_server", {})
_webserver_settings = settings.get("web_server", {})
WEBSERVER_HOST = _webserver_settings.get("host", "0.0.0.0")
WEBSERVER_PORT = int(_webserver_settings.get("port", 5000))

######################
#    ANALYTICS      #
######################

ANALYTICS = settings.get("analytics", None)

####################
# INSTRUMENTATION  #
####################

INSTRUMENTATION = settings.get("instrumentation", {})
INSTRUMENTATION_ENABLED = INSTRUMENTATION.get("enabled", False)

####################
# Enabled ADAPTERS #
####################
ADAPTERS = set()

for adapter in settings.get("adapters", set()):
    ADAPTERS.add(settings["adapters"][adapter].get("kind"))
